#include "GLWidget.hpp"

#include <QTimer>
#include <iostream>
#include <il.h>
#include <io.h>

GLWidget::GLWidget(QWidget *parent)
: QGLWidget(QGLFormat(QGL::SampleBuffers), parent)
{
	setMinimumWidth(768);
	setMinimumHeight(768);

	data = new FlowData;

	curChannel = 0;

	showStreamlines = false;
	euler = true;
	glyph = false;
	dsep = 1.0f;
	dtest = 0.5f * dsep;
	dt = 1.0f;

	rS = gS = bS = aS = 1.0f;
	rA = gA = bA = aA = 1.0f;

	showArrows = false;
	numArrows = 5;
}

GLWidget::~GLWidget()
{
	for(int i=0; i < data->getNumUsedChannels(); i++)
	{
		glDeleteTextures(1, &texChannels[i]);
	}

	glDeleteTextures(1, &texTransfer);
	glDeleteTextures(1, &texArrow);
}


void GLWidget::initializeGL()
{
	// Initialize GLEW
	std::cout << "- Initializing GLEW ..." << std::endl;
	GLenum err = glewInit();
	if (err != GLEW_OK)
	{
		// glewInit failed, something is seriously wrong
		std::cerr << "Error initializing GLEW: " << glewGetErrorString(err) << std::endl;
		std::cerr << EXIT_FAILURE << std::endl;
	}
	std::cout << "- GLEW initialized." << std::endl << std::endl;

	// Initialize DevIL (an image library)
	std::cout << "- Initializing DevIL ..." << std::endl;
	ilInit();
	if (ilGetError() != IL_NO_ERROR)
	{
		std::cerr << "Error initializing DevIL" << std::endl;
	}
	ilEnable(IL_ORIGIN_SET);
	ilOriginFunc(IL_ORIGIN_LOWER_LEFT);
	std::cout << "- DevIL initialized." << std::endl << std::endl;

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	glEnable(GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable(GL_POINT_SPRITE);
	glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
	glTexEnvf(GL_POINT_SPRITE_ARB,
		GL_COORD_REPLACE_ARB, GL_TRUE);

	glEnable(GL_TEXTURE_1D);
	glEnable(GL_TEXTURE_2D);

	glGenTextures(1, &texTransfer);
	glGenTextures(1, &texArrow);

	// SHADER
	shader = new QGLShaderProgram();

	if(!shader->addShaderFromSourceFile(QGLShader::Vertex, "res/flow.vert"))
		printf(shader->log().toAscii());
	if(!shader->addShaderFromSourceFile(QGLShader::Fragment, "res/flow.frag"))
		printf(shader->log().toAscii());

	wchar_t buff[100];
	swprintf(buff, L"Getting GLSL files..\n", shader->log().toAscii());
	LPCWSTR p = buff;
	OutputDebugString(p);

	shader->link();


	// SHADER
	arrowShader = new QGLShaderProgram();

	if(!arrowShader->addShaderFromSourceFile(QGLShader::Vertex, "res/arrow.vert"))
		printf(shader->log().toAscii());
	if(!arrowShader->addShaderFromSourceFile(QGLShader::Fragment, "res/arrow.frag"))
		printf(shader->log().toAscii());

	swprintf(buff, L"Getting GLSL files..\n", shader->log().toAscii());
	p = buff;
	OutputDebugString(p);

	arrowShader->link();


	// load arrow.png
	ILuint img;
	ilGenImages(1, &img);
	ilBindImage(img);
	if(!ilLoadImage("res/arrow.png"))
	{
		wchar_t buff[100];
		swprintf(buff, L"Error loading image!\n");
		LPCWSTR p = buff;
		OutputDebugString(p);
		return;
	}
	if(!ilConvertImage(IL_RGBA, GL_UNSIGNED_BYTE))
	{
		wchar_t buff[100];
		swprintf(buff, L"Error converting image!\n");
		LPCWSTR p = buff;
		OutputDebugString(p);
		return;
	}
	glBindTexture(GL_TEXTURE_2D, texArrow);
	glTexImage2D(GL_TEXTURE_2D, 0,
		ilGetInteger(IL_IMAGE_BPP),
		ilGetInteger(IL_IMAGE_WIDTH), ilGetInteger(IL_IMAGE_HEIGHT),
		0, ilGetInteger(IL_IMAGE_FORMAT),
		GL_UNSIGNED_BYTE, ilGetData());
	glTexParameteri(GL_TEXTURE_2D,
		GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,
		GL_TEXTURE_MIN_FILTER, GL_LINEAR);
}

void GLWidget::resizeGL(int width, int height)
{
	resize(width, height);

	int side = (width < height) ? width : height;

	glViewport(
		(width == side) ? 0 : (width - side) / 2,
		(height == side) ? 0 : (height - side) / 2,
		side,
		side);

	update();
}

void GLWidget::paintGL()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	if(!data->isLoaded())
		return;

	float w = (geometry->getDimX() > geometry->getDimY())
		? 1.0f : (float) geometry->getDimX() / geometry->getDimY();
	float h = (geometry->getDimX() < geometry->getDimY())
		? 1.0f : (float) geometry->getDimY() / geometry->getDimX();
	glScalef(2.0f * w, 2.0f * h, 0.0f);
	glTranslatef(-0.5f, -0.5f, 0.0f);

	if(!data->isLoaded())
		return;
	// draw rectangels

	int geoDimX = geometry->getDimX();
	int geoDimY = geometry->getDimY();

	vec3 geoSize = geometry->getPosXY(geoDimX - 1,
		geoDimY - 1);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texChannels[curChannel]);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_1D, texTransfer);

	if(!shader->bind())
	{
		wchar_t buff[100];
		swprintf(buff, L"Error: %s\n", shader->log().toAscii());
		LPCWSTR p = buff;
		OutputDebugString(p);
	}

	shader->setUniformValue("flowData", GLuint(0));
	shader->setUniformValue("transferData", GLuint(1));

	glBegin(GL_QUADS);
	
	for(int x = 0; x < geoDimX - 1; x++)
	{
		for(int y = 0; y < geoDimY - 1; y++)
		{
			vec3 v = geometry->getPosXY(x, y);
			glTexCoord2f((float) x / (geoDimX - 1),
				(float) y / (geoDimY - 1));
			glVertex2f(v[0] / geoSize[0],
				v[1] / geoSize[1]);
			
			v = geometry->getPosXY(x+1, y);
			glTexCoord2f((float) (x+1) / (geoDimX - 1),
				(float) y / (geoDimY - 1));
			glVertex2f(v[0] / geoSize[0],
				v[1] / geoSize[1]);
			
			v = geometry->getPosXY(x+1, y+1);
			glTexCoord2f((float) (x+1) / (geoDimX - 1),
				(float) (y+1) / (geoDimY - 1));
			glVertex2f(v[0] / geoSize[0],
				v[1] / geoSize[1]);
			
			v = geometry->getPosXY(x, y+1);
			glTexCoord2f((float) x / (geoDimX - 1),
				(float) (y+1) / (geoDimY - 1));
			glVertex2f(v[0] / geoSize[0],
				v[1] / geoSize[1]);
		}
	}

	glEnd();

	shader->release();

	if(showStreamlines)
	{
		if(lines.size() < 1)
			renderStreamlines();

		for(int i=0; i < lines.size(); i++)
		{
			glColor4f(rS, gS, bS, aS);

			if(!glyph)
			{
				glBegin(GL_LINE_STRIP);
				bool valid;
				vec3 pos = lines[i].getNextCandidate(&valid);
				while((valid))
				{
					pos = geometry->normalizeCoords(pos);
					glVertex2f(pos[0], pos[1]);
					pos = lines[i].getNextCandidate(&valid);
				}
				glEnd();
			}
			else
			{
				glActiveTexture(GL_TEXTURE2);
				glBindTexture(GL_TEXTURE_2D, texArrow);

				arrowShader->bind();
				arrowShader->setUniformValue("arrowData", GLuint(2));
				glColor4f(rS, gS, bS, aS);

				glBegin(GL_POINTS);
				bool valid;
				vec3 pos = lines[i].getNextCandidate(&valid);
				while((valid))
				{
					vec3 vec(
					(data->getChannel(0)->getValue(pos) - data->getChannel(0)->getMin())
						/ data->getChannel(0)->getRange(),
					(data->getChannel(1)->getValue(pos) - data->getChannel(1)->getMin())
						/ data->getChannel(1)->getRange());

					//vec = !vec;
					glTexCoord3f(vec[0], vec[1], -dt);
					
					pos = geometry->normalizeCoords(pos);
					glVertex2f(pos[0], pos[1]);

					pos = lines[i].getNextCandidate(&valid);
				}
				glEnd();
			}
		}
	}

	if(showArrows)
		renderArrows();
	glFlush();
}

void GLWidget::renderArrows()
{
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_2D, texArrow);

	arrowShader->bind();
	arrowShader->setUniformValue("arrowData", GLuint(2));

	glColor4f(rA, gA, bA, aA);
	glBegin(GL_POINTS);

	float step = 1.0f / (float)numArrows;

	for(float x = step/2.0f; x < 1.0f; x+=step)
	{
		for(float y = step/2.0f; y < 1.0f; y+=step)
		{
			vec3 pos = geometry->unNormalizeCoords(vec3(x, y));
			glTexCoord3f(
				(data->getChannel(0)->getValue(pos) - data->getChannel(0)->getMin())
					/ data->getChannel(0)->getRange(),
				(data->getChannel(1)->getValue(pos) - data->getChannel(1)->getMin())
					/ data->getChannel(1)->getRange(),
					20.0f);
			glVertex2f(x, y);
		}
	}

	glEnd();
}

void GLWidget::renderStreamlines()
{
	// compute initial streamline and put it into the queue
	// calculate initial streamline
	generateStreamline(geometry->unNormalizeCoords(vec3(0.5f, 0.5f)));

	// let this initial streamline be the current streamline
	int linePos = 0;

	bool finished = false;

	do
	{
		vec3 candidate1;
		vec3 candidate2;
		bool validCandidate1 = true;
		bool validCandidate2 = true;

		do
		{
			bool valid;
			validCandidate1 = validCandidate2 = true;
			vec3 addr = lines[linePos].getNextCandidate(&valid);
			if(!valid) {
				validCandidate1 = validCandidate2 = false;
				break;
			}

			vec3 vector(data->getChannel(0)->getValue(addr),
				data->getChannel(1)->getValue(addr));

			if(vector == vec3(0.0f, 0.0f)) {
				validCandidate1 = validCandidate2 = false;
				break;
			}

			vector = !vector;
			vector = vec3(vector[1], -vector[0]);

			candidate1 = addr + vector * dsep;
			candidate2 = addr - vector * dsep;

			if(candidate1[0] < geometry->getMinX()
					|| candidate1[0] > geometry->getMaxX()
					|| candidate1[1] < geometry->getMinY()
					|| candidate1[1] > geometry->getMaxY())
					validCandidate1 = false;
			if(candidate2[0] < geometry->getMinX()
					|| candidate2[0] > geometry->getMaxX()
					|| candidate2[1] < geometry->getMinY()
					|| candidate2[1] > geometry->getMaxY())
					validCandidate2 = false;

			for(int i = 0; i < lines.size(); i++)
			{
				if(validCandidate1)
					validCandidate1 = lines[i].checkPoint(candidate1, dtest, 0);
				if(validCandidate2)
					validCandidate2 = lines[i].checkPoint(candidate2, dtest, 0);

				if(!validCandidate1 && !validCandidate2)
					break;
			}

		} while(!validCandidate1 && !validCandidate2);
		
		if(validCandidate1 || validCandidate2) //valid candidate
		{
			//compute a new streamline and put it into the queue
			if(validCandidate1)
				generateStreamline(candidate1);
			if(validCandidate2)
				generateStreamline(candidate2);
		}
		else
		{
			linePos++;
			if(linePos >= lines.size()) // there is no more available streamline in the queue
				finished = true;
		}
	} while(!finished);

	update();
}

void GLWidget::generateStreamline(vec3 seedingPoint)
{
	vec3 currentPos = seedingPoint;

	Streamline currentLine;
	currentLine.addPointBack(currentPos);

	do
	{
		vec3 tmp = currentPos;
		if(euler)
			if((currentPos -= calcEulerPos(currentPos)) == tmp)
				break;
			else;
		else
			if((currentPos -= calcRungeKuttaPos(currentPos)) == tmp)
				break;

		// check other stream lines position
		bool valid = true;
		for(int i = 0; i < lines.size(); i++)
		{
			if((valid = lines[i].checkPoint(currentPos, dtest, 0)) != true)
				break;
		}
		if(!valid)
			break;

		int delta = 1;
		if(dt < dtest)
			delta += dtest/dt;
		if(!currentLine.checkPoint(currentPos, dtest, delta))
			break;

		currentLine.addPointBack(currentPos);

	} while(true);

	lines.push_back(currentLine);

	currentLine = *new Streamline;

	currentLine.addPointBack(seedingPoint);
	currentPos = seedingPoint;

	do
	{
		vec3 tmp = currentPos;
		if(euler)
			if((currentPos += calcEulerPos(currentPos)) == tmp)
				break;
			else;
		else
			if((currentPos += calcRungeKuttaPos(currentPos)) == tmp)
				break;

		// check other stream lines position
		bool valid = true;
		for(int i = 0; i < lines.size() - 1; i++)
		{
			if((valid = lines[i].checkPoint(currentPos, dtest, 0)) != true)
				break;
		}
		if(!valid)
			break;

		int delta = 1;
		if(dt < dtest)
			delta += dtest/dt;
		if(!currentLine.checkPoint(currentPos, dtest, delta))
			break;
		if(!lines[lines.size()-1].checkPoint(currentPos, dtest, delta))
			break;

		currentLine.addPointBack(currentPos);

	} while(true);

	lines.push_back(currentLine);

	update();
}

vec3 GLWidget::calcEulerPos(vec3 pos)
{
	if(pos[0] < geometry->getMinX()
		|| pos[0] > geometry->getMaxX()
		|| pos[1] < geometry->getMinY()
		|| pos[1] > geometry->getMaxY())
		return vec3(0.0f, 0.0f);

	vec3 nextPos(
		data->getChannel(0)->getValue(pos),
		data->getChannel(1)->getValue(pos));
	if(nextPos == vec3(0.0f, 0.0f))
		return nextPos;

	nextPos = !nextPos;

	pos += nextPos * dt;

	if(pos[0] < geometry->getMinX()
		|| pos[0] > geometry->getMaxX()
		|| pos[1] < geometry->getMinY()
		|| pos[1] > geometry->getMaxY())
		return vec3(0.0f, 0.0f);

	return nextPos * dt;
}

vec3 GLWidget::calcRungeKuttaPos(vec3 pos)
{
	if(pos.v[0] < geometry->getMinX()
		|| pos.v[0] > geometry->getMaxX()
		|| pos.v[1] < geometry->getMinY()
		|| pos.v[1] > geometry->getMaxY())
		return vec3(0.0f, 0.0f);

	vec3 nextPos(
		data->getChannel(0)->getValue(pos),
		data->getChannel(1)->getValue(pos));
	if(nextPos == vec3(0.0f, 0.0f))
		return nextPos;

	nextPos = !nextPos;

	nextPos = pos + nextPos * dt/2.0f;

	if(nextPos[0] < geometry->getMinX()
		|| nextPos[0] > geometry->getMaxX()
		|| nextPos[1] < geometry->getMinY()
		|| nextPos[1] > geometry->getMaxY())
		return vec3(0.0f, 0.0f);

	nextPos = vec3(data->getChannel(0)->getValue(nextPos),
			data->getChannel(1)->getValue(nextPos));

	if(nextPos == vec3(0.0f, 0.0f))
		return nextPos;
	
	nextPos = !nextPos;
	pos += (nextPos * dt);

	if(pos[0] < geometry->getMinX()
		|| pos[0] > geometry->getMaxX()
		|| pos[1] < geometry->getMinY()
		|| pos[1] > geometry->getMaxY())
		return vec3(0.0f, 0.0f);

	return nextPos * dt;
}


void GLWidget::loadData(const std::string & strFilename)
{
	// delete old data
	
	for(int i=0; i < data->getNumUsedChannels(); i++)
	{
		glDeleteTextures(1, &texChannels[i]);
	}


	if(!data->loadDataset(strFilename, false))
	{
		wchar_t buff[100];
		swprintf(buff, L"Error loading data!\n");
		LPCWSTR p = buff;
		OutputDebugString(p);
		return;
	}

	geometry = data->getGeometry();

	wchar_t buff[100];
	swprintf(buff, L"Data loaded successfully!\n");
	LPCWSTR p = buff;
	OutputDebugString(p);


	// prepare data

	int geoDimX = geometry->getDimX();
	int geoDimY = geometry->getDimY();

	for(int i = 0; i < data->getNumUsedChannels()-3; i++)
	{
		float *channelData = new float[geoDimX * geoDimY];

		for(int x = 0; x < geoDimX; x++)
		{
			for(int y = 0; y < geoDimY; y++)
			{
				channelData[y * geoDimX + x] =
					(data->getChannel(i+3)->getValue(
						geometry->getVtx(x, y))
						- data->getChannel(i+3)->getMin()) /
					data->getChannel(i+3)->getRange();
			}
		}

		// create textures

		glGenTextures(1, & texChannels[i+1]);
		glBindTexture(GL_TEXTURE_2D, texChannels[i+1]);
		glTexImage2D(GL_TEXTURE_2D,
			0, GL_LUMINANCE,
			geoDimX,
			geoDimY,
			0, GL_LUMINANCE, GL_FLOAT, channelData);
		glTexParameteri(GL_TEXTURE_2D,
				GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,
				GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		delete[] channelData;
	}

	float *channelData = new float[geoDimX * geoDimY];

	for(int x = 0; x < geoDimX; x++)
	{
		for(int y = 0; y < geoDimY; y++)
		{
			vec3 vec((data->getChannel(0)->getValue(geometry->getVtx(x, y))
					- data->getChannel(0)->getMin()) /
					data->getChannel(0)->getRange(),
				(data->getChannel(1)->getValue(geometry->getVtx(x, y))
					- data->getChannel(1)->getMin()) /
					data->getChannel(1)->getRange());

			channelData[y * geoDimX + x] = vec.length();
		}
	}

	// create textures

	glGenTextures(1, & texChannels[0]);
	glBindTexture(GL_TEXTURE_2D, texChannels[0]);
	glTexImage2D(GL_TEXTURE_2D,
		0, GL_LUMINANCE,
		geoDimX,
		geoDimY,
		0, GL_LUMINANCE, GL_FLOAT, channelData);
	glTexParameteri(GL_TEXTURE_2D,
			GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,
			GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	delete[] channelData;

	if(showStreamlines)
		renderStreamlines();

	update();

	emit numChannelsChanged(data->getNumUsedChannels() - 2);
}


void GLWidget::saveScreenshot()
{
	unsigned int uImage = 0;
	std::vector<unsigned char> vecImage;

	char vcFilename[1024];
	// Check for next free filename, since we don't want to overwrite existing images
	// images will be saved as 'VisLuFramework0000.png','VisLuFramework0001.png','VisLuFramework0002.png',etc.
	for (int i=0;i<9999;i++)
	{
		sprintf_s(vcFilename,1024,"VisLuFramework%04d.png",i);

		// check if file does not exist
		if( (_access(vcFilename,0)) == -1 )
			break;
	}

	std::cout << "- Saving screenshot to '" << vcFilename << "' ..." << std::endl;

	vecImage.resize(width()*height()*4);

	// Read the OpenGL framebuffer
	glPushAttrib(GL_PIXEL_MODE_BIT);
	glReadBuffer(GL_FRONT);

	glReadPixels(0,0,width(),height(),GL_RGBA,GL_UNSIGNED_BYTE,(void*) &(vecImage.front()) );
	const GLenum glError = glGetError();	
	
	glPopAttrib();

	if (glError != GL_NO_ERROR)
	{
		std::cerr<<"Unable to read frame buffer "<<gluErrorString(glError)<<std::endl;
	}
	else
	{
		ilGenImages(1, &uImage);

		if (ilGetError() != IL_NO_ERROR)
		{
			std::cerr<<"Unable to create image"<<std::endl;
		}
		else
		{
			ilBindImage(uImage);

			if (ilGetError() != IL_NO_ERROR)
			{
				std::cerr<<"Unable to bind image"<<std::endl;
			}
			else
			{
				ilTexImage(width(),height(),1,4,IL_RGBA,IL_UNSIGNED_BYTE,(void*) &(vecImage.front()) );

				if (ilGetError() != IL_NO_ERROR)
				{
					std::cerr<<"Unable to store image"<<std::endl;
				}
				else
				{
					// Use DevIL to save a screenshot of the OpenGL window
					ilSaveImage(vcFilename);

					if (ilGetError() != IL_NO_ERROR)
					{
						std::cerr<<"Unable to save image"<<std::endl;
					}
					else
					{
						std::cout << "- Screenshot saved." << std::endl;
					}
				}
			}
		}
	}
	if (uImage)
	{
		ilBindImage(0);
		ilDeleteImages(1,&uImage);
	}
}

// slots
void GLWidget::setTransferValues(float *_values)
{
	glBindTexture(GL_TEXTURE_1D, texTransfer);

	glTexImage1D(GL_TEXTURE_1D, 0,
		GL_RGBA, 4096, 0,
		GL_RGBA, GL_FLOAT,
		_values);

	glTexParameteri(GL_TEXTURE_1D,
			GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_1D,
			GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	delete[] _values;

	update();
}

void GLWidget::channelChanged(int _channel)
{
	curChannel = _channel;
	update();
}

void GLWidget::switchStreamlineState(bool _showStreamlines)
{
	showStreamlines = _showStreamlines;
	update();
}

void GLWidget::dtValueChanged(double _dt)
{
	lines.clear();
	dt = (float)_dt;
	if(showStreamlines)
		renderStreamlines();
	update();
}

void GLWidget::dsepValueChanged(double _dsep)
{
	lines.clear();
	dtest /= dsep;
	dsep = (float)_dsep;
	dtest = dsep * dtest;
	if(showStreamlines)
		renderStreamlines();
	update();
}

void GLWidget::dtestValueChanged(double _dtest)
{
	lines.clear();
	dtest = (float)_dtest * dsep;
	if(showStreamlines)
		renderStreamlines();
	update();
}

void GLWidget::switchArrowState(bool _showArrows)
{
	showArrows = _showArrows;
	update();
}

void GLWidget::numArrowChanged(int _numArrows)
{
	numArrows = _numArrows;
	update();
}

void GLWidget::interpolationMethodChanged(bool _euler)
{
	lines.clear();
	euler = _euler;
	if(showStreamlines)
		renderStreamlines();
	update();
}

void GLWidget::paintMethodChanged(bool _glyph)
{
	glyph = _glyph;
	update();
}

void GLWidget::alphaStreamChanged(int alpha)
{
	aS = (float) alpha / 10.0f;
	update();
}

void GLWidget::alphaArrowChanged(int alpha)
{
	aA = (float) alpha / 10.0f;
	update();
}


void GLWidget::setArrowColor(float r, float g, float b)
{
	rA = r;
	gA = g;
	bA = b;
	update();
}

void GLWidget::setStreamColor(float r, float g, float b)
{
	rS = r;
	gS = g;
	bS = b;
	update();
}