#ifndef GLWIDGET_HPP_
#define GLWIDGET_HPP_

#include <GL/glew.h>

#include <QGLWidget>
#include <QtOpenGL>
#include <QGLShaderProgram>
#include <QGLShader>
#include <QColor>

#include <vector>

#include "FlowData.h"
#include "FlowGeometry.h"
#include "Streamline.h"

/**
 * Class which represents the OpenGL rendering window.
 * In this Class all needed functions for handling the volume and slice rendering are implemented.
 * @author Christian Hirsch
 * @author Clemens Roegner
 */
class GLWidget : public QGLWidget
{
	Q_OBJECT

private:
	bool showArrows;
	bool showStreamlines;

	FlowData *data;
	FlowGeometry *geometry;

	GLuint texTransfer, texArrow;
	GLuint texChannels[max_channels];
	QGLShaderProgram *shader, *arrowShader;

	int curChannel;
	int xRot, yRot, zRot;
	QPoint lastPos;

	vector<Streamline> lines;
	bool euler;
	bool glyph;
	float dsep;
	float dtest;
	float dt;

	float rA,gA,bA,aA;
	float rS,gS,bS,aS;

	int numArrows;

protected:
	/**
	 * Initializes all data needed for rendering (including devil + glut).
	 */
	void initializeGL();
	/**
	 * Initializes the OpenGL render window when the window is resized.
	 * @param width The new width of the render window.
	 * @param height The new height of the render window.
	 */
	void resizeGL(int width, int height);
	/**
	 * Render function. Renders either the slices or the full volume data.
	 */
    void paintGL();
	/**
	 * Arrow render function.
	 */
	void renderArrows();
	/**
	 * Streamline render funtcion. This function calculates new streamlines with
	 * the alogrithm to generate "Evenly-Spaced Streamlines of Arbitrary Density".
	 */
	void renderStreamlines();
	/**
	 * Generate streamline function. This function creates one streamline.
	 * @param seedingPoint The point of which a streamline should be created.
	 */
	void generateStreamline(vec3 seedingPoint);

	/**
	 * Calculates the next position with the Euler interpolation Method.
	 * @param pos The current Position.
	 * @return The vector to the next position.
	 */
	vec3 calcEulerPos(vec3 pos);
	/**
	 * Calculates the next position with the Runge-Kutta interpolation Method.
	 * @param pos The current Position.
	 * @return The vector to the next position.
	 */
	vec3 calcRungeKuttaPos(vec3 pos);

public:
	/**
	 * Constructor.
	 * This constructor creates a new OpenGL render window.
	 * @param parent Parent QWidget element.
	 */
	GLWidget(QWidget *parent = 0);
	/**
	 * Destructor.
	 * Releases all the needed (graphics) memory.
	 */
	virtual ~GLWidget();
	/**
	 * Loads new flow data.
	 * @param strFilename String to the path of the volume data.
	 */
	void loadData(const std::string & strFilename);

	/**
	 * Sets new values for the transferfunction.
	 * @param *_values Pointer to the first element of a 4096 element float array.
	 */
	void setTransferValues(float *_values);

	/**
	 * Changes the colors of the arrows.
	 * @param r The new red value.
	 * @param g The new green value.
	 * @param b The new blue value.
	 */
	void setArrowColor(float r, float g, float b);

	/**
	 * Changes the colors of the streams.
	 * @param r The new red value.
	 * @param g The new green value.
	 * @param b The new blue value.
	 */
	void setStreamColor(float r, float g, float b);

public slots:
	/**
	 * This function saves a screenhot in the programs folder.
	 * The file name is created automatically.
	 */
	void saveScreenshot();

	/**
	 * Function to change the displayed channel.
	 * @param index Index of the channel.
	 */
	void channelChanged(int index);

	/**
	 * Changes the streamline state to [not] draw the streamlines.
	 * @param state True if streamlines should be drawn.
	 */
	void switchStreamlineState(bool state);
	/**
	 * Changes dt value for the interpolation methods.
	 * @param dt New dt value.
	 */
	void dtValueChanged(double dt);
	/**
	 * Changes seperation value for the streamlines.
	 * @param dt New dsep value.
	 */
	void dsepValueChanged(double);
	/**
	 * Changes test value for the streamlines.
	 * @param dt New dtest value.
	 */
	void dtestValueChanged(double);
	/**
	 * Changes the interpolation method for the streamlines.
	 * @param euler If true, then the Euler method will be used, else the RK method is used.
	 */
	void interpolationMethodChanged(bool euler);
	/**
	 * Changes the paint method (normal lines vs glyph mapping).
	 * @param glyph If true the glyph maps are drawn, else normal lines.
	 */
	void paintMethodChanged(bool glyph);

	/**
	 * Changes the arrow state to [not] draw the arrows.
	 * @param state True if arrows should be drawn.
	 */
	void switchArrowState(bool state);
	/**
	 * Changes the number of arrows which should be drawn.
	 * @param numArrows The number of arrows in one line and of lines. numArrows * numArrows of arrows will be darwn.
	 */
	void numArrowChanged(int glyph);
	/**
	 * Changes the alpha value of the arrows.
	 * @param alpha The new alpha value.
	 */
	void alphaArrowChanged(int alpha);
	/**
	 * Changes the alpha value of the arrows.
	 * @param alpha The new alpha value.
	 */
	void alphaStreamChanged(int alpha);


signals:

	/**
	 * Signal which is emitted when the number of channels has changed.
	 * @param channels The new number of channels.
	 */
	void numChannelsChanged(int channels);
};

#endif /* GLWIDGET_HPP_ */
