#include <QGraphicsRectItem>
#include <QMouseEvent>
#include <stdio.h>

#include "TransferWidget.hpp"
#include "TransferPoint.hpp"

TransferWidget::TransferWidget(QWidget *parent = 0)
: QGraphicsView(parent)
{
	// turn off scroll bars
	setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
	setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

	// scene initialization
	scene = new QGraphicsScene(this);
	setScene(scene);

	TransferPoint *pnt = new TransferPoint(0, 1);
	pnt->setFixed(true);
	pnt->setColor(Qt::black);
	scene->addItem(pnt);
	points.append(pnt);

	pnt = new TransferPoint(1, 0);
	pnt->setFixed(true);
	pnt->setColor(Qt::white);
	scene->addItem(pnt);
	points.append(pnt);

	QGraphicsLineItem *line = new QGraphicsLineItem(
			points[0]->x(), points[0]->y(),
			points[1]->x(), points[1]->y());
	lines.append(line);
	scene->addItem(line);

	setRenderHint(QPainter::Antialiasing);
	setBackgroundBrush(QPixmap("res/background.png"));
	setCacheMode(QGraphicsView::CacheBackground);

	fillBackground();
}

void TransferWidget::mousePressEvent(QMouseEvent *event)
{
	QGraphicsView::mousePressEvent(event);

	if(event->button() == Qt::RightButton)
	{
		TransferPoint *pnt = new TransferPoint(
			event->posF().x() / scene->width(),
			event->posF().y() / scene->height());
		pnt->setColorFromDialog();
		scene->addItem(pnt);

		// put to correct position
		int i;
		for(i=0; points[i]->getRelativeX() < pnt->getRelativeX(); i++)
			;
		points.insert(i, pnt);
	}
	scene->removeItem(background);
	redrawScene();
}

void TransferWidget::mouseMoveEvent(QMouseEvent *event)
{
	QGraphicsView::mouseMoveEvent(event);

	redrawScene();
}

void TransferWidget::resizeEvent(QResizeEvent *event)
{
	scene->setSceneRect(0, 0, width()-2, height()-2);
	redrawScene();
}

void TransferWidget::redrawScene()
{
	points[0]->updateScene();
	// update positions of transfer points
	for(int i=1; i<points.size(); i++)
	{
		points[i]->updateScene();

		// swap points
		if(points[i-1]->x() > points[i]->x())
			points.swap(i-1, i);
	}
	scene->removeItem(background);
	fillBackground();
}

void TransferWidget::fillBackground()
{
	QLinearGradient gradient(scene->sceneRect().bottomLeft(), scene->sceneRect().bottomRight());
	QPainterPath *bg = new QPainterPath(scene->sceneRect().bottomLeft());

	for(int i=0; i<points.size(); i++)
	{
		gradient.setColorAt(points[i]->getRelativeX(), points[i]->getColor());
		bg->lineTo(points[i]->pos());
	}
	bg->lineTo(scene->sceneRect().bottomRight());
	bg->lineTo(scene->sceneRect().bottomLeft());

	background = scene->addPath(*bg, QPen(Qt::black, 1), QBrush(gradient));

	emit transferChanged();
}

float *TransferWidget::getTransferValues()
{
	float *transferValues = new float[4096*4];

	for(int i=1; i<points.size(); i++)
	{
		int start = points[i-1]->getRelativeX()*4096;
		int end = points[i]->getRelativeX()*4096;

		float deltaR = points[i]->getColor().redF() - points[i-1]->getColor().redF();
		deltaR /= (end-start);
		float deltaG = points[i]->getColor().greenF() - points[i-1]->getColor().greenF();
		deltaG /= (end-start);
		float deltaB = points[i]->getColor().blueF() - points[i-1]->getColor().blueF();
		deltaB /= (end-start);
		float deltaA = points[i]->getColor().alphaF() - points[i-1]->getColor().alphaF();
		deltaA /= (end-start);

		// red
		transferValues[start*4 + 0] = points[i-1]->getColor().redF();
		// green
		transferValues[start*4 + 1] = points[i-1]->getColor().greenF();
		// blue
		transferValues[start*4 + 2] = points[i-1]->getColor().blueF();
		//alpha
		transferValues[start*4 + 3] = points[i-1]->getColor().alphaF();

		for(int j = start+1; j < end; j++)
		{
			// red
			transferValues[j*4 + 0] = transferValues[(j-1)*4 + 0] + deltaR;
			// green
			transferValues[j*4 + 1] = transferValues[(j-1)*4 + 1] + deltaG;
			// blue
			transferValues[j*4 + 2] = transferValues[(j-1)*4 + 2] + deltaB;
			//alpha
			transferValues[j*4 + 3] = transferValues[(j-1)*4 + 3] + deltaA;

			//printf("%f/%f/%f/%f\n", transferValues[j*4 + 0], transferValues[j*4 + 1], transferValues[j*4 + 2], transferValues[j*4 + 3]);
		}
	}
	//printf("%f/%f/%f/%f\n", transferValues[4096*4-4], transferValues[4096*4-3], transferValues[4096*4-2], transferValues[4096*4-1]);
	return transferValues;
}

float *TransferWidget::getTransferControlValues()
{
	float *transferValues = new float[points.size() * 5];

	for(int i=0; i<points.size(); i++)
	{
		// red
		transferValues[i*5 + 0] = points[i]->getColor().redF();
		// green
		transferValues[i*5 + 1] = points[i]->getColor().greenF();
		// blue
		transferValues[i*5 + 2] = points[i]->getColor().blueF();

		// x/y
		transferValues[i*5 + 3] = (float)points[i]->getRelativeX();
		transferValues[i*5 + 4] = (float)points[i]->getRelativeY();
	}
	return transferValues;
}

int TransferWidget::getNumPoints()
{
	return points.size();
}

void TransferWidget::clear()
{
	points.clear();

	scene->clear();

	// scene initialization
	scene = new QGraphicsScene(this);
	setScene(scene);

	fillBackground();
}

void TransferWidget::setNewRGBXY(float _red, float _green, float _blue, float _x, float _y)
{
	QColor *color = new QColor;
	
	color->setRedF(_red);
	color->setGreenF(_green);
	color->setBlueF(_blue);
	color->setAlphaF(1.0);

	TransferPoint *pnt = new TransferPoint(_x, _y);
	pnt->setColor(*color);
	scene->addItem(pnt);
	points.append(pnt);

	//redrawScene();
}